@@ -81,14 +81,6 @@ kbnc_z (double *k, unsigned long *b, unsigned long *n, signed long *c, mpz_t z)
   if (k == NULL || b == NULL || n == NULL || c == NULL)
     return 0;
 
-  /* make sure the input meets some sort of minimum size requirement.
-     The gwnum library reports ES1_CANNOT_DO_QUICKLY for number < 2^350 */
-  if (mpz_sizeinbase(z, 2) < 350)
-  {
-    *b = 0;
-    return 0;
-  }
-
   mpz_init (diff);
   mpz_init (abs_diff);
   mpz_init (b_n);
@@ -99,8 +91,11 @@ kbnc_z (double *k, unsigned long *b, unsigned long *n, signed long *c, mpz_t z)
   mpz_init (base);
   mpz_init_set_ui (base_min, 2);
   mpz_init_set_ui (base_max, 10000);
-  /* this puts a bound on how large of a k value we want to find */
-  mpz_init_set_str (max_k, "562949953421312", 10);
+
+  /* this puts a bound on how large of a k value we want to find (2^49-1) */
+  mpz_init_set_ui (max_k, 2);
+  mpz_pow_ui (max_k, max_k, 49);
+  mpz_sub_ui (max_k, max_k, 1);
 
   /* when dividing: z/(base^exp) this will give us a possible k value */
   /* we want a quick test to see if this might be a viable k value */
@@ -209,22 +204,32 @@ kbnc_z (double *k, unsigned long *b, unsigned long *n, signed long *c, mpz_t z)
 }
 
 /* This function searches for a nice representation of z
-   We are trying to see if z = k*b^n + c
+   We are trying to see if z = (k*b^n + c)/(factors, if any)
    Some examples that we can find:
    "3^218+5123"
    "(199*3^218+5123)/(2*17*587*1187)"
-   "(199*3^218 + 5123)/2/17/587/1187" */
+   "(199*3^218 + 5123)/2/17/587/1187"
+
+   This function also handles inputs of the form
+   z = (k*b^(b2^n2) + c)/(factors, if any) */
 int
 kbnc_str (double *k, unsigned long *b, unsigned long *n, signed long *c,
           char *z, mpz_t num)
 {
-  unsigned long i = 0;
-  int total = 0;
-  char strk[11];
+  unsigned long i = 0, s_len, b2, n2, i2, pow;
+  unsigned long prev_pwr_indx = 0, last_pwr_indx = 0,
+                mul_indx = 0, sign_indx = 0;
+  int total = 0, power_count = 0, mul_count = 0, sign_count = 0;
+  char strk[20];
   char strb[11];
   char strn[11];
   char strc[11];
+  char strb2[11];
+  char strn2[11];
+  char sign=0;
   mpz_t tmp;
+  uint64_t test1, test2;
+  unsigned long klen;
 
   /* make sure we have a place to put our results */
   if (k == NULL || b == NULL || n == NULL || c == NULL || z == NULL)
@@ -232,17 +237,74 @@ kbnc_str (double *k, unsigned long *b, unsigned long *n, signed long *c,
 
   *b = 0;
 
-  for (i = 0; i < strlen(z); i++)
+  /* ignore everything after a '/', if there is one */
+  s_len = strlen(z);
+  for (i = 0; i < s_len; i++)
+  {
+    if( z[i] == '^' )
+    {
+      power_count++;
+      prev_pwr_indx = last_pwr_indx;
+      last_pwr_indx = i;
+    }
+
+    if( z[i] == '*' )
+    {
+      mul_count++;
+      mul_indx = i;
+    }
+
+    if( z[i] == '+' )
+    {
+      sign_count++;
+      sign_indx = i;
+      sign = 1;
+    }
+
+    if( z[i] == '-' )
+    {
+      sign_count++;
+      sign_indx = i;
+      sign = -1;
+    }
+
+    if( z[i] == '/' )
+    {
+       s_len = i;
+       break;
+    }
+  }
+  /* exit if there is no '^' in the string
+     or the input is not of the right form */
+  if( power_count == 0 || power_count > 2 || mul_count > 1 || sign_count != 1 ||
+        sign_indx < last_pwr_indx || mul_indx > sign_indx)
+    return 0;
+
+  if( power_count == 1 && mul_indx < last_pwr_indx)
+  {
+    for (i = 0; i < s_len; i++)
     {
       if (z[i] == '(' || z[i] == '{' || z[i] == '[')
         continue;
 
       /* check to see if the input is k*b^n+c */
-      total = sscanf (z+i, "%10[0-9]*%10[0-9]^%10[0-9]%*[ +]%10[0-9]",
+      total = sscanf (z+i, "%17[0-9]*%10[0-9]^%10[0-9]%*[ +]%10[0-9]",
                       strk, strb, strn, strc);
       if (total == 4)
         {
-          *k = (double) strtoul (strk, NULL, 10);
+          klen = strlen(strk);
+          if( klen > 10 )
+          {
+            test2 = strtoul (strk+10, NULL, 10);
+            strk[10] = 0;
+            test1 = strtoul (strk, NULL, 10);
+            for( i = 0; i < (klen - 10); i++)
+              test1 *= 10;
+            *k = (double)(test1 + test2); 
+          }
+          else
+            *k = (double) strtoul (strk, NULL, 10);
+
           *b = strtoul (strb, NULL, 10);
           *n = strtoul (strn, NULL, 10);
           *c = strtol (strc, NULL, 10);
@@ -250,11 +312,23 @@ kbnc_str (double *k, unsigned long *b, unsigned long *n, signed long *c,
         }
 
       /* check to see if the input is k*b^n-c */
-      total = sscanf (z+i, "%10[0-9]*%10[0-9]^%10[0-9]%*[ -]%10[0-9]",
+      total = sscanf (z+i, "%17[0-9]*%10[0-9]^%10[0-9]%*[ -]%10[0-9]",
                       strk, strb, strn, strc);
       if (total == 4)
         {
-          *k = (double) strtoul (strk, NULL, 10);
+          klen = strlen(strk);
+          if( klen > 10 )
+          {
+            test2 = strtoul (strk+10, NULL, 10);
+            strk[10] = 0;
+            test1 = strtoul (strk, NULL, 10);
+            for( i = 0; i < (klen - 10); i++)
+              test1 *= 10;
+            *k = (double)(test1 + test2); 
+          }
+          else
+            *k = (double) strtoul (strk, NULL, 10);
+
           *b = strtoul (strb, NULL, 10);
           *n = strtoul (strn, NULL, 10);
           *c = strtol (strc, NULL, 10);
@@ -288,6 +362,94 @@ kbnc_str (double *k, unsigned long *b, unsigned long *n, signed long *c,
         }
       break;
     }
+  }
+  else if( power_count == 2 && mul_indx < prev_pwr_indx )
+  {
+    for (i = 0; i < prev_pwr_indx; i++)
+    {
+      if (z[i] == '(' || z[i] == '{' || z[i] == '[')
+        continue;
+
+      /* find k & b */
+      if( mul_indx == 0 ) /* k = 1.0 */
+      {
+        total = sscanf (z+i, "%10[0-9]", strb);
+        if(total == 1)
+        {
+          *k = 1.0;
+          *b = strtoul (strb, NULL, 10);
+          break;
+        }
+      }
+      else
+      {
+        total = sscanf (z+i, "%17[0-9]*%10[0-9]", strk, strb);
+        if( total == 2 )
+        {
+          klen = strlen(strk);
+          if( klen > 10 )
+          {
+            test2 = strtoul (strk+10, NULL, 10);
+            strk[10] = 0;
+            test1 = strtoul (strk, NULL, 10);
+            for( i = 0; i < (klen - 10); i++)
+              test1 *= 10;
+            *k = (double)(test1 + test2); 
+          }
+          else
+            *k = (double) strtoul (strk, NULL, 10);
+
+          *b = strtoul (strb, NULL, 10);
+          break;
+        }
+      }
+      break;
+    }
+
+    for (i = (prev_pwr_indx + 1); i < s_len; i++)
+    {
+      if (z[i] == '(' || z[i] == '{' || z[i] == '[')
+        continue;
+
+      total = sscanf (z+i, "%10[0-9]^%10[0-9]", strb2, strn2);
+      if (total == 2)
+      {
+        b2 = strtoul (strb2, NULL, 10);
+        n2 = strtoul (strn2, NULL, 10);
+
+        if( (double)n2*log10((double)b2) >= 10.0 )
+        {
+          outputf (OUTPUT_NORMAL, 
+           "Exponent (%lu^%lu) too large in kbnc_str!\n", b2, n2);
+          return 0;
+        }
+
+        pow = b2;
+        for( i2 = 1; i2 < n2; i2++)
+          pow *= b2;
+
+        *n = pow;
+        break;
+      }
+      break;
+    }
+
+    for (i = (sign_indx + 1); i < s_len; i++)
+    {
+      if (z[i] == '(' || z[i] == '{' || z[i] == '[')
+        continue;
+
+      total = sscanf (z+i, "%10[0-9]", strc);
+      if(total == 1)
+      {
+        *c = sign*strtoul (strc, NULL, 10);
+        break;
+      }
+      break;
+    }
+  }
+  else
+    return 0;
 
   /* first, check to see if we found a k*b^n+c */
   if (*b)
@@ -315,29 +477,21 @@ kbnc_str (double *k, unsigned long *b, unsigned long *n, signed long *c,
   return 0;
 }
 
-/* return ceil(log(v)/log(2)) */
-unsigned long gw_log_2(unsigned long v)
-{
-  unsigned long r = 0;
-
-  while (v > 1)
-    {
-      r++;
-      v = (v + 1) / 2;
-    }
-
-  return r;
-}
-
 int 
 gw_ecm_stage1 (mpz_t f, curve *P, mpmod_t modulus, 
 	       double B1, double *B1done, mpz_t go, double gw_k,
                unsigned long gw_b, unsigned long gw_n, signed long gw_c)
 {
   ecm_uint gw_B1done = *B1done;
-  unsigned long siz_x, siz_z; /* Size of gw_x and gw_y as longs */
-  mpz_t gw_x, gw_z, gw_A;
+  unsigned long siz_x, siz_z, kbnc_size, options; /* Size of gw_x and gw_y as longs */
+  double tmp_bitsize;
+  mpz_t gw_x, gw_z, gw_A, tmp;
   int youpi;
+  char gwnum_msg[8][24];
+
+  /* P->y must never be zero when calling ecm_mul */
+  if( P->y->_mp_size == 0 )
+    mpres_set_ui (P->y, 1, modulus);
 
   if (mpz_cmp_ui (go, 1) > 0)
     {
@@ -349,17 +503,53 @@ gw_ecm_stage1 (mpz_t f, curve *P, mpmod_t modulus,
       mpres_clear (b, modulus);
     }
   
-  outputf (OUTPUT_VERBOSE, 
+  if(gw_b)
+  {
+    outputf (OUTPUT_NORMAL, 
            "Using gwnum_ecmStage1(%.0f, %d, %d, %d, %.0f, %ld)\n",
            gw_k, gw_b, gw_n, gw_c, B1, gw_B1done);
 
-  /* Copy x, z and A values from modular representation to 
-     plain integers */
+    /* make sure tmp has adequate allocation */
+    tmp_bitsize = log2(gw_k) + ((double)gw_n)*log2((double)gw_b) + 64.0;
+    tmp_bitsize = 64.0*ceil(tmp_bitsize/64.0); /* set to first multiple of 64 >= tmp_bitsize */
+    mpz_init2 (tmp, (unsigned long)tmp_bitsize);
+  
+    /* construct k*b^n+c to get true size */
+    mpz_set_ui (tmp, gw_b);
+    mpz_pow_ui (tmp, tmp, gw_n);
+    mpz_mul_ui (tmp, tmp, (unsigned long)gw_k);
+    if (gw_c >= 0)
+      mpz_add_ui (tmp, tmp, gw_c);
+    else
+      mpz_sub_ui (tmp, tmp, (gw_c * -1));
 
-  /* Allocate enough memory for any residue (mod k*b^n+c) for x, z */
-  mpz_init2 (gw_x, (gw_n+1)*gw_log_2(gw_b)+64);
-  mpz_init2 (gw_z, (gw_n+1)*gw_log_2(gw_b)+64);
-  mpz_init (gw_A);
+    /* kbnc_size = bits per word * # of whole words required to hold k*b^n+c */
+    kbnc_size = 8*sizeof(mp_size_t)*(tmp->_mp_size); 
+    ASSERT_ALWAYS ( (unsigned long)tmp_bitsize >= kbnc_size );
+    mpz_clear (tmp);
+
+    /* Allocate enough memory for any residue (mod k*b^n+c) for x, z */
+    /* ecmstag1.c in gwnum says it needs 60 bits more than the gwnum modulus size,
+       so we add 64 bits here to maintain whole-word allocations for gw_x and gw_z */
+    mpz_init2 (gw_x, kbnc_size+64);
+    mpz_init2 (gw_z, kbnc_size+64);
+    mpres_init (gw_A, modulus);
+    options = 1; /* set flag for gwnum to do "slow" jobs; users can now test using the
+                     -force-gwnum and -force-no-gwnum command line flags which way
+                      is actually faster for a given input */
+  }
+  else /* set for gwnum generic mod */
+  {
+    outputf (OUTPUT_NORMAL, 
+           "Using gwnum_ecmStage1_generic(%.0f, %ld)\n", B1, gw_B1done);
+
+     kbnc_size = 8*sizeof(mp_size_t)*ABSIZ(modulus->orig_modulus) + 64; /* One extra 64-bit word per G. Woltman */
+     mpz_init2 (gw_x, kbnc_size);
+     mpz_init2 (gw_z, kbnc_size);
+     mpres_init (gw_A, modulus);
+     gw_k = 1.0; gw_n = 1; gw_c = 1;
+     options = 1; /* set flag for gwnum to do "slow" jobs; George W. has improved generic reduction (MMGW algorithm) */
+  }
 
   /* mpres_get_z always produces non-negative integers */
   mpres_get_z (gw_x, P->x, modulus);
@@ -371,21 +561,23 @@ gw_ecm_stage1 (mpz_t f, curve *P, mpmod_t modulus,
   siz_x = SIZ(gw_x);
   siz_z = SIZ(gw_z);
 
-  /* George Woltman says that the gwnum library can handle k values up to 49
-     or 50 bits long, and the maximum c value is +/-8388607 */
+  /* Per p95v3019b11.source/gwnum/gwnum.c, line 2181:
+  "if (k > 9007199254740991.0) return (GWERROR_K_TOO_LARGE);"
+  The stated value is 2^53-1, so we must require k <= 2^53-1.
+  The maximum c value is +/-8388607 (2^23-1) */
   ASSERT_ALWAYS (gw_k == rint (gw_k)); /* check that k is an integer */
-  ASSERT_ALWAYS (1.0 <= gw_k && gw_k <= 562949953421312.0);
+  ASSERT_ALWAYS (1.0 <= gw_k && gw_k < 0x1p53);
   ASSERT_ALWAYS (-8388607 <= gw_c && gw_c <= 8388607);
 #if GMP_NUMB_BITS <= 32
   youpi = gwnum_ecmStage1_u32 (gw_k, gw_b, gw_n, gw_c, 
       PTR(modulus->orig_modulus), ABSIZ(modulus->orig_modulus), 
       B1, &gw_B1done, PTR(gw_A), ABSIZ(gw_A), 
-      PTR(gw_x), &siz_x, PTR(gw_z), &siz_z, NULL, 0);
+      PTR(gw_x), &siz_x, PTR(gw_z), &siz_z, NULL, options);
 #else /* contributed by David Cleaver */
   youpi = gwnum_ecmStage1_u64 (gw_k, gw_b, gw_n, gw_c,
       PTR(modulus->orig_modulus), ABSIZ(modulus->orig_modulus),
       B1, &gw_B1done, PTR(gw_A), ABSIZ(gw_A),
-      PTR(gw_x), &siz_x, PTR(gw_z), &siz_z, NULL, 0);
+      PTR(gw_x), &siz_x, PTR(gw_z), &siz_z, NULL, options);
 #endif
 
   /* Test that not more was written to gw_x and gw_z than we had space for */
@@ -418,19 +610,31 @@ gw_ecm_stage1 (mpz_t f, curve *P, mpmod_t modulus,
 /* #define ES1_MEMORY            3 *//* Out of memory */
 /* #define ES1_INTERRUPT         4 *//* Execution interrupted */
 /* #define ES1_CANNOT_DO_QUICKLY 5 *//* Requires 3-multiply reduction */
-/* #define ES1_HARDWARE_ERROR    6 *//* An error was detected, most */
+/* #define ES1_HARDWARE_ERROR    6 *//* An error was detected, most likely a hardware error. */
+
+  strcpy( gwnum_msg[0], "ES1_SUCCESS");
+  strcpy( gwnum_msg[1], "ES1_FACTOR_FOUND");
+  strcpy( gwnum_msg[2], "ES1_CANNOT_DO_IT");
+  strcpy( gwnum_msg[3], "ES1_MEMORY");
+  strcpy( gwnum_msg[4], "ES1_INTERRUPT");
+  strcpy( gwnum_msg[5], "ES1_CANNOT_DO_QUICKLY");
+  strcpy( gwnum_msg[6], "ES1_HARDWARE_ERROR");
+
   if (youpi == ES1_CANNOT_DO_IT || youpi == ES1_CANNOT_DO_QUICKLY)
   {
     outputf (OUTPUT_VERBOSE, 
            "Notice: Did not use gwnum_ecmStage1(%.0f, %d, %d, %d, %.0f, %ld)\n",
            gw_k, gw_b, gw_n, gw_c, B1, gw_B1done);
+    outputf (OUTPUT_VERBOSE, "Reason message: %s\n", gwnum_msg[youpi]);
     youpi = ECM_NO_FACTOR_FOUND;
     goto end_of_gwecm;
   }
 
   if (youpi > 1)
     {
-      outputf (OUTPUT_ERROR, "GW stage 1 returned code %d\n", youpi);
+      outputf (OUTPUT_ERROR, "GW stage 1 returned error code %d\n", youpi);
+      outputf (OUTPUT_VERBOSE, "GW stage 1 returned error code %d\n", youpi);
+      outputf (OUTPUT_VERBOSE, "Reason message: %s\n", gwnum_msg[youpi]);
       youpi = ECM_ERROR;
       goto end_of_gwecm;
     }
@@ -440,6 +644,7 @@ gw_ecm_stage1 (mpz_t f, curve *P, mpmod_t modulus,
       /* How did that happen? Since we passed z, GWNUM should not do
          an extgcd and so not find factors... but if it did anyways, 
          we deal with it. Who's going to turn down a factor? */
+      /* Note (PBM): gwnum will report a factor if gcd(gw_A, orig_modulus) > 1 */
       outputf (OUTPUT_DEVVERBOSE, 
                "gw_ecm_stage1: Strange, gwnum_ecmStage1 reports a factor\n");
       mpres_get_z (f, P->x, modulus);
