@@ -3,42 +3,12 @@
 #ifdef WITH_GPU
 
 #include "cudacommon.h"
-#include "cudakernel.h"
 
-#ifdef HAVE_CGBN_H
 #include "cgbn_stage1.h"
-#endif /* HAVE_CGBN_H */
 
 
 #define TWO32 4294967296 /* 2^32 */
 
-int findfactor (mpz_t factor, mpz_t N, mpz_t xfin, mpz_t zfin)
-{
-  int youpi;
-  mpz_t gcd;
-  mpz_init (gcd);
-
-  mpz_gcd (gcd, zfin, N);
-  
-  if (mpz_cmp_ui (gcd, 1) == 0)
-  {
-    mpz_invert (zfin, zfin, N);
-    mpz_mul (xfin, xfin, zfin);
-    mpz_mod (xfin, xfin, N);
-      
-    mpz_set (factor, xfin);
-    youpi = ECM_NO_FACTOR_FOUND;
-  }
-  else //gcd !=1 (and gcd>0 because N>0) so we found a factor
-  {
-      mpz_set (factor, gcd);
-      youpi = ECM_FACTOR_FOUND_STEP1;
-    }
-
-  mpz_clear (gcd);
-  return youpi;
-}
-
 /* Try to reduce all composite factors to primes.
  * This can be hard if factors overlap e.g. (a*b, a*c*d, b*c)
  */
@@ -171,56 +141,14 @@ void reducefactors (mpz_t *factors, int *array_found, unsigned int nb_curves)
 }
 
 
-void to_mont_repr (mpz_t x, mpz_t n)
-{
-  mpz_mul_2exp (x, x, ECM_GPU_MAX_BITS);
-  mpz_mod (x, x, n);
-}
-
-void from_mont_repr (mpz_t x, mpz_t n, mpz_t invB)
-{
-  mpz_mul (x, x, invB);
-  mpz_mod (x, x, n);
-}
-
-void mpz_to_biguint (biguint_t a, mpz_t b)
-{
-  int i;
-
-  for (i=0;i<ECM_GPU_NB_DIGITS;i++)
-  {
-#if GMP_NUMB_BITS == 32
-    a[i]=mpz_getlimbn (b, i);  
-#else // GMP_NUMB_BITS == 64
-    if (i%2 == 0)
-      a[i]=(mpz_getlimbn (b, i/2) & 0x00000000ffffffff);
-    else
-      a[i]=(mpz_getlimbn (b, i/2) >> 32);  
-#endif
-  }
-}
-
-void biguint_to_mpz (mpz_t a, biguint_t b)
-{
-  int i;
-  
-  mpz_set_ui (a, 0);
-
-  for (i=ECM_GPU_NB_DIGITS-1;i>=0;i--)
-  {
-    mpz_mul_2exp (a, a, 32);
-	  mpz_add_ui (a , a, b[i]);
-  }
-}
-
 static void
 A_from_sigma (mpz_t A, unsigned int sigma, mpz_t n)
 {
   mpz_t tmp;
   int i;
   mpz_init_set_ui (tmp, sigma);
-  /* Compute d = sigma/2^ECM_GPU_SIZE_DIGIT */
-  for (i = 0; i < ECM_GPU_SIZE_DIGIT; i++)
+  /* Compute d = sigma/2^32 */
+  for (i = 0; i < 32; i++)
     {
       if (mpz_tstbit (tmp, 0) == 1)
       mpz_add (tmp, tmp, n);
@@ -235,149 +163,8 @@ A_from_sigma (mpz_t A, unsigned int sigma, mpz_t n)
 }
 
 
-int gpu_ecm_stage1 (mpz_t *factors, int *array_found, mpz_t N, mpz_t s,
-                    unsigned int number_of_curves, unsigned int firstsigma, 
-                    float *gputime, int verbose)
-{
-  int youpi = ECM_NO_FACTOR_FOUND;
-
-  unsigned int sigma;
-  unsigned int i;
-
-  mpz_t N3; /* N3 = 3*N */
-  mpz_t w; /* w = 2^(SIZE_DIGIT) */
-  mpz_t invN; /* invN = -N^-1 mod w */
-  mpz_t invB; /* invB = 2^(-MAX_BITS) mod N ; B is w^NB_DIGITS */
-  mpz_t invw; /* w^(-1) mod N */
-  mpz_t M; /* (invN*N+1)/w */
-  mpz_t xp, zp, x2p, z2p;
-
-  /* The same variables but for the GPU */
-  biguint_t *h_xarray, *h_zarray, *h_x2array, *h_z2array;
-  digit_t h_invN;
-  biguint_t h_N, h_3N, h_M;
-
-  /*****************************/
-  /* Initialize some variables */
-  /*****************************/
-  mpz_init (N3);
-  mpz_init (invw);
-  mpz_init (w);
-  mpz_init (M);
-  mpz_init (xp);
-  mpz_init (zp);
-  mpz_init (x2p);
-  mpz_init (z2p);
-  mpz_init (invN);
-  mpz_init (invB);
-
-  h_xarray= (biguint_t *) malloc (number_of_curves * sizeof (biguint_t));
-  h_zarray= (biguint_t *) malloc (number_of_curves * sizeof (biguint_t));
-  h_x2array= (biguint_t *) malloc (number_of_curves * sizeof (biguint_t));
-  h_z2array= (biguint_t *) malloc (number_of_curves * sizeof (biguint_t));
-
-  /*Some computation depending on N */
-  mpz_mul_ui (N3, N, 3); /* Compute N3 = 3*N */
-  mpz_ui_pow_ui (w, 2, ECM_GPU_SIZE_DIGIT); /* Compute w = 2^SIZE_DIGIT */
-    
-  mpz_invert (invN, N, w);
-  mpz_sub (invN, w, invN); /* Compute invN = -N^-1 mod w */
-   
-  mpz_mul (M, invN, N);
-  mpz_add_ui (M, M, 1);
-  mpz_divexact (M, M, w); /* Compute M = (invN*N+1)/w */
-
-  mpz_to_biguint (h_N, N); 
-  mpz_to_biguint (h_3N, N3); 
-  mpz_to_biguint (h_M, M); 
-  h_invN = mpz_get_ui (invN); 
-   
-  mpz_ui_pow_ui (invB, 2, ECM_GPU_MAX_BITS); 
-  mpz_invert (invB, invB, N); /* Compute invB = 2^(-MAX_BITS) mod N */
-
-  mpz_invert (invw, w, N); /* Compute inw = 2^-SIZE_DIGIT % N */
-
-  /* xp zp x2p are independent of N and the curve */
-  mpz_set_ui (xp, 2);
-  mpz_set_ui (zp, 1);
-  mpz_set_ui (x2p, 9);
-
-  /* Compute their Montgomery representation */
-  to_mont_repr (xp, N);
-  to_mont_repr (zp, N);
-  to_mont_repr (x2p, N);
-  
-  /* for each curve, compute z2p and put xp, zp, x2p, z2p in the h_*array  */
-  for (i = 0; i < number_of_curves; i++)
-  {
-    sigma = firstsigma + i;
-
-    mpz_mul_ui (z2p, invw, sigma);
-    mpz_mod (z2p, z2p, N);
-    mpz_mul_2exp (z2p, z2p, 6);
-    mpz_add_ui (z2p, z2p, 8);
-    mpz_mod (z2p, z2p, N); /* z2p = 8+64*d */
-
-    to_mont_repr (z2p, N);
-
-    mpz_to_biguint (h_xarray[i], xp); 
-    mpz_to_biguint (h_zarray[i], zp); 
-    mpz_to_biguint (h_x2array[i], x2p); 
-    mpz_to_biguint (h_z2array[i], z2p); 
-  } 
- 
-  /* Call the wrapper function that call the GPU */
-  *gputime=cuda_Main (h_N, h_3N, h_M, h_invN, h_xarray, h_zarray, h_x2array, 
-                     h_z2array, s, firstsigma, number_of_curves, verbose);
-
-  /* Analyse results */
-  for (i = 0; i < number_of_curves; i++)
-  {
-    sigma = firstsigma + i;
-
-    biguint_to_mpz (xp, h_xarray[i]); 
-    biguint_to_mpz (zp, h_zarray[i]); 
-    
-    from_mont_repr (xp, N, invB);
-    from_mont_repr (zp, N, invB);
-  
-    array_found[i] = findfactor (factors[i], N, xp, zp);
-
-    if (array_found[i] != ECM_NO_FACTOR_FOUND)
-      {
-        youpi = array_found[i];
-        outputf (OUTPUT_NORMAL, "GPU: factor %Zd found in Step 1 with"
-                " curve %u (-sigma 3:%u)\n", factors[i], i, sigma);
-      }
-  }
-  
-  mpz_clear (N3);
-  mpz_clear (invN);
-  mpz_clear (invw);
-  mpz_clear (w);
-  mpz_clear (M);
-  mpz_clear (xp);
-  mpz_clear (zp);
-  mpz_clear (x2p);
-  mpz_clear (z2p);
-  mpz_clear (invB);
-  
-  free ((void *) h_xarray);
-  free ((void *) h_zarray);
-  free ((void *) h_x2array);
-  free ((void *) h_z2array);
-
-  return youpi;
-}
-
 int
-gpu_ecm (mpz_t f, mpz_t x, int param, mpz_t firstsigma, mpz_t n, mpz_t go,
-         double *B1done, double B1, mpz_t B2min_parm, mpz_t B2_parm, 
-         unsigned long k, const int S, int verbose, int repr,
-         int nobase2step2, int use_ntt, int sigma_is_A, FILE *os, FILE* es, 
-         char *chkfilename ATTRIBUTE_UNUSED, char *TreeFilename, double maxmem,
-         int (*stop_asap)(void), mpz_t batch_s, double *batch_last_B1_used, 
-         int use_cgbn, int device, int *device_init, unsigned int *nb_curves)
+gpu_ecm (mpz_t f, const ecm_params params, ecm_params mutable_params, mpz_t n, double B1)
 {
   unsigned int i;
   int youpi = ECM_NO_FACTOR_FOUND;
@@ -400,18 +187,21 @@ gpu_ecm (mpz_t f, mpz_t x, int param, mpz_t firstsigma, mpz_t n, mpz_t go,
   mpz_t B2min, B2; /* Local B2, B2min to avoid changing caller's values */
   unsigned long dF;
   root_params_t root_params;
+  unsigned int nb_curves = 0; /* Local copy of number of curves */
 
-  ASSERT((-1 <= sigma_is_A) && (sigma_is_A <= 1));
+  /* This helps keep track of when params is changed in this function. */
+  assert((void*) params == (void*) mutable_params); /* params != mutable params */
+
+  ASSERT((-1 <= params->sigma_is_A) && (params->sigma_is_A <= 1));
   ASSERT((GMP_NUMB_BITS == 32) || (GMP_NUMB_BITS == 64));
 
   /* Set global VERBOSE to avoid the need to explicitly passing verbose */
-  set_verbose (verbose);
-  ECM_STDOUT = (os == NULL) ? stdout : os;
-  ECM_STDERR = (es == NULL) ? stdout : es;
-
+  set_verbose (params->verbose);
+  ECM_STDOUT = (params->os == NULL) ? stdout : params->os;
+  ECM_STDERR = (params->es == NULL) ? stdout : params->es;
 
   /* Check that N is not too big */
-  size_t max_bits = (use_cgbn ? ECM_GPU_CGBN_MAX_BITS : ECM_GPU_MAX_BITS) - 6;
+  size_t max_bits = ECM_GPU_CGBN_MAX_BITS - 6;
   if (mpz_sizeinbase (n, 2) > max_bits)
     {
       outputf (OUTPUT_ERROR, "GPU: Error, input number should be stricly lower"
@@ -420,26 +210,46 @@ gpu_ecm (mpz_t f, mpz_t x, int param, mpz_t firstsigma, mpz_t n, mpz_t go,
     }
 
   /* Only param = ECM_PARAM_BATCH_32BITS_D is accepted on GPU */
-  if (param == ECM_PARAM_DEFAULT)
-      param = ECM_PARAM_BATCH_32BITS_D;
-    
-  if (param != ECM_PARAM_BATCH_32BITS_D)
+  if (params->param == ECM_PARAM_DEFAULT)
+      mutable_params->param = ECM_PARAM_BATCH_32BITS_D;
+
+  if (params->param != ECM_PARAM_BATCH_32BITS_D)
     {
       outputf (OUTPUT_ERROR, "GPU: Error, only param = ECM_PARAM_BATCH_32BITS_D "
                              "is accepted on GPU.\n");
       return ECM_ERROR;
     }
 
+  if (params->go != NULL && mpz_cmp_ui (params->go, 1) > 0)
+    {
+      outputf (OUTPUT_ERROR, "GPU: Error, option -go is not allowed\n");
+      return ECM_ERROR;
+    }
+
+  /* Cannot do resume on GPU */
+  if (!ECM_IS_DEFAULT_B1_DONE(params->B1done) && params->B1done < B1)
+    {
+      outputf (OUTPUT_ERROR, "GPU: Error, cannot resume on GPU.\n");
+      return ECM_ERROR;
+    }
+
+  /* Current code works only for sigma_is_A = 0 */
+  if (params->sigma_is_A != 0)
+    {
+      outputf (OUTPUT_ERROR, "GPU: ERROR, sigma_is_A not allowed.\n");
+      return ECM_ERROR;
+    }
+
   /* check that repr == ECM_MOD_DEFAULT or ECM_MOD_BASE2 (only for stage 2) */
-  if (repr != ECM_MOD_DEFAULT && repr != ECM_MOD_BASE2)
+  if (params->repr != ECM_MOD_DEFAULT && params->repr != ECM_MOD_BASE2)
       outputf (OUTPUT_ERROR, "GPU: Warning, the value of repr will be ignored "
       "for step 1 on GPU.\n");
 
   /* It is only for stage 2, it is not taken into account for GPU code */
-  if (mpmod_init (modulus, n, repr) != 0)
+  if (mpmod_init (modulus, n, params->repr) != 0)
     return ECM_ERROR;
 
-  /* See what kind of number we have as that may influence optimal parameter 
+  /* See what kind of number we have as that may influence optimal parameter
      selection. Test for base 2 number. Note: this was already done by
      mpmod_init. */
 
@@ -448,33 +258,13 @@ gpu_ecm (mpz_t f, mpz_t x, int param, mpz_t firstsigma, mpz_t n, mpz_t go,
 
   /* For a Fermat number (base2 a positive power of 2) */
   for (Fermat = base2; Fermat > 0 && (Fermat & 1) == 0; Fermat >>= 1);
-  if (Fermat == 1) 
+  if (Fermat == 1)
     {
       Fermat = base2;
       po2 = 1;
     }
   else
       Fermat = 0;
- 
-  /* Cannot do resume on GPU */
-  if (!ECM_IS_DEFAULT_B1_DONE(*B1done) && *B1done < B1)
-    {
-      outputf (OUTPUT_ERROR, "GPU: Error, cannot resume on GPU.\n");
-      return ECM_ERROR;
-    }
-
-  /* Compute s */
-  if (B1 != *batch_last_B1_used || mpz_cmp_ui (batch_s, 1) <= 0)
-    {
-      *batch_last_B1_used = B1;
-
-      st = cputime ();
-      /* construct the batch exponent */
-      compute_s (batch_s, B1, NULL);
-      outputf (OUTPUT_VERBOSE, "Computing batch product (of %" PRIu64
-                               " bits) of primes up to B1=%1.0f took %ldms\n",
-                               mpz_sizeinbase (batch_s, 2), B1, cputime () - st);
-    }
 
   /* Set parameters for stage 2 */
   mpres_init (P.x, modulus);
@@ -484,87 +274,70 @@ gpu_ecm (mpz_t f, mpz_t x, int param, mpz_t firstsigma, mpz_t n, mpz_t go,
   mpz_init (B2);
   mpz_init (B2min);
 
-  youpi = set_stage_2_params (B2, B2_parm, B2min, B2min_parm, &root_params,
-                              B1, &k, S, use_ntt, &po2, &dF,
-                              TreeFilename, maxmem, Fermat, modulus);
+  youpi = set_stage_2_params (
+          B2, params->B2,
+          B2min, params->B2min,
+          &root_params,
+          B1, &mutable_params->k, params->S, params->use_ntt, &po2, &dF,
+                              params->TreeFilename, params->maxmem, Fermat, modulus);
   if (youpi == ECM_ERROR)
       goto end_gpu_ecm;
 
-  /* Set cudaDeviceScheduleBlockingSync with -cgbn, else cudaDeviceScheduleYield */
-  int schedule = use_cgbn ? 1 : 0;
-
   /* Initialize the GPU if necessary and determine nb_curves */
-  if (!*device_init)
+  if (!params->gpu_device_init)
     {
       st = cputime ();
-      youpi = select_and_init_GPU (device, nb_curves,
-                                   test_verbose (OUTPUT_VERBOSE), schedule);
+      youpi = select_and_init_GPU (
+          params->gpu_device,
+          &mutable_params->gpu_number_of_curves,
+          test_verbose (OUTPUT_VERBOSE));
 
       if (youpi != 0)
         {
           youpi = ECM_ERROR;
-          goto end_gpu_ecm2;
+          goto end_gpu_ecm;
         }
 
       outputf (OUTPUT_VERBOSE, "GPU: Selection and initialization of the device "
                                "took %ldms\n", elltime (st, cputime ()));
       /* TRICKS: If initialization of the device is too long (few seconds), */
       /* try running 'nvidia-smi -q -l' on the background .                 */
-      *device_init = 1;
+      mutable_params->gpu_device_init = 1;
     }
 
-  /* Init arrays */
-  factors = (mpz_t *) malloc (*nb_curves * sizeof (mpz_t));
-  ASSERT_ALWAYS (factors != NULL);
+  /* Number of curves is only set after select_and_init_GPU */
+  nb_curves = params->gpu_number_of_curves;
 
-  array_found = (int *) malloc (*nb_curves * sizeof (int));
-  ASSERT_ALWAYS (array_found != NULL);
-
-  for (i = 0; i < *nb_curves; i++)
-    {
-      mpz_init (factors[i]);
-      array_found[i] = ECM_NO_FACTOR_FOUND;
-    }
-
-
-  /* Current code works only for sigma_is_A = 0 */
-  if (sigma_is_A != 0)
-    {
-      outputf (OUTPUT_ERROR, "GPU: Not yet implemented.\n");
-      youpi= ECM_ERROR;
-      goto end_gpu_ecm;
-    }
-
-  ASSERT (sigma_is_A == 0);
-  if (mpz_sgn (firstsigma) == 0)
+  ASSERT (params->sigma_is_A == 0);
+  if (mpz_sgn (params->sigma) == 0)
     {
       /* generate random value in [2, 2^32 - nb_curves - 1] */
-      mpz_set_ui (firstsigma, (get_random_ul () %
-                               (TWO32 - 2 - *nb_curves)) + 2);
+      mpz_set_ui (mutable_params->sigma,
+                  (get_random_ul () % (TWO32 - 2 - nb_curves)) + 2);
     }
   else /* sigma should be in [2, 2^32-nb_curves] */
     {
-      if (mpz_cmp_ui (firstsigma, 2) < 0 || 
-          mpz_cmp_ui (firstsigma, TWO32 - *nb_curves) >= 0)
+      if (mpz_cmp_ui (params->sigma, 2) < 0 ||
+          mpz_cmp_ui (params->sigma, TWO32 - nb_curves) >= 0)
         {
           outputf (OUTPUT_ERROR, "GPU: Error, sigma should be in [2,%lu]\n",
-                                 TWO32 - *nb_curves - 1);
-          youpi= ECM_ERROR;
+                                 TWO32 - nb_curves - 1);
+          youpi = ECM_ERROR;
           goto end_gpu_ecm;
         }
     }
-  firstsigma_ui = mpz_get_ui (firstsigma);
+  firstsigma_ui = mpz_get_ui (params->sigma);
 
-  print_B1_B2_poly (OUTPUT_NORMAL, ECM_ECM, B1, *B1done,  B2min_parm, B2min,
-                    B2, S, firstsigma, sigma_is_A, ECM_EC_TYPE_MONTGOMERY,
-                    go, param, *nb_curves);
-  outputf (OUTPUT_VERBOSE, "dF=%lu, k=%lu, d=%lu, d2=%lu, i0=%Zd\n", 
-           dF, k, root_params.d1, root_params.d2, root_params.i0);
+  print_B1_B2_poly (OUTPUT_NORMAL, ECM_ECM, B1, params->B1done,  params->B2min, B2min,
+                    B2, params->S, params->sigma, params->sigma_is_A, ECM_EC_TYPE_MONTGOMERY,
+                    params->go, params->param, nb_curves);
+  outputf (OUTPUT_VERBOSE, "dF=%lu, k=%lu, d=%lu, d2=%lu, i0=%Zd\n",
+           dF, params->k, root_params.d1, root_params.d2, root_params.i0);
 
-  if (go != NULL && mpz_cmp_ui (go, 1) > 0)
+  if ((mpz_cmp (B2, B2min) < 0) && modulus->repr == ECM_MOD_BASE2 && params->nobase2step2)
     {
-      outputf (OUTPUT_ERROR, "GPU: Error, option -go is not allowed\n");
-      youpi= ECM_ERROR;
+      outputf (OUTPUT_ERROR, "GPU: ERROR, -nobase2s2 not supported on GPU.\n");
+      youpi = ECM_ERROR;
       goto end_gpu_ecm;
     }
 
@@ -572,95 +345,100 @@ gpu_ecm (mpz_t f, mpz_t x, int param, mpz_t firstsigma, mpz_t n, mpz_t go,
     {
       if (mpz_cmp_d (B2min, B1) != 0)
         {
-          outputf (OUTPUT_VERBOSE, 
+          outputf (OUTPUT_VERBOSE,
             "Can't compute success probabilities for B1 <> B2min\n");
         }
       else
         {
           rhoinit (256, 10);
-          print_expcurves (B1, B2, dF, k, root_params.S, param);
+          print_expcurves (B1, B2, dF, params->k, root_params.S, params->param);
         }
     }
 
+  /* Init arrays */
+  factors = (mpz_t *) malloc (nb_curves * sizeof (mpz_t));
+  ASSERT_ALWAYS (factors != NULL);
+
+  array_found = (int *) malloc (nb_curves * sizeof (int));
+  ASSERT_ALWAYS (array_found != NULL);
+
+  for (i = 0; i < nb_curves; i++)
+    {
+      mpz_init (factors[i]);
+      array_found[i] = ECM_NO_FACTOR_FOUND;
+    }
+
+
+  /* Compute s */
+  if (B1 != params->batch_last_B1_used || mpz_cmp_ui (params->batch_s, 1) <= 0)
+    {
+      mutable_params->batch_last_B1_used = B1;
+
+      st = cputime ();
+      /* construct the batch exponent */
+      compute_s (mutable_params->batch_s, B1, NULL);
+      outputf (OUTPUT_VERBOSE, "Computing batch product (of %" PRIu64
+                               " bits) of primes up to B1=%1.0f took %ldms\n",
+                               mpz_sizeinbase (params->batch_s, 2), B1, cputime () - st);
+    }
+
+
   st = cputime ();
 
-  if (use_cgbn) {
-#ifdef HAVE_CGBN_H
-    youpi = cgbn_ecm_stage1 (factors, array_found, n, batch_s, *nb_curves,
-                             firstsigma_ui, &gputime, verbose);
-#else
-    outputf (OUTPUT_ERROR, "cgbn not included");
-    return ECM_ERROR;
-#endif /* HAVE_CGBN_H */
-  }  else {
-    youpi = gpu_ecm_stage1 (factors, array_found, n, batch_s, *nb_curves,
-                            firstsigma_ui, &gputime, verbose);
-  }
+  youpi = cgbn_ecm_stage1 (factors, array_found, n, params->batch_s, nb_curves,
+                           firstsigma_ui, &gputime, params->verbose);
 
   outputf (OUTPUT_NORMAL, "Computing %u Step 1 took %ldms of CPU time / "
-                          "%.0fms of GPU time\n", *nb_curves, 
-                                           elltime (st, cputime ()), gputime);
-  outputf (OUTPUT_VERBOSE, "Throughput: %.3f curves per second ", 
-                                                 1000 * (*nb_curves)/gputime);
-  outputf (OUTPUT_VERBOSE, "(on average %.2fms per Step 1)\n", 
-                                                        gputime/(*nb_curves));
+                          "%.0fms of GPU time\n", nb_curves,
+                          elltime (st, cputime ()), gputime);
+  outputf (OUTPUT_VERBOSE, "Throughput: %.3f curves per second ",
+                           1000 * nb_curves/gputime);
+  outputf (OUTPUT_VERBOSE, "(on average %.2fms per Step 1)\n",
+                           gputime/nb_curves);
   tottime = (long) gputime;
 
-  *B1done=B1;
+  mutable_params->B1done = B1;
 
   /* GMP documentation says mpz_sizeinbase(op, 2) is always the exact value. */
   size_t n_bits = mpz_sizeinbase(n, 2);
 
   /* Save stage 1 residues as x = x0 + x1 * 2^bits + ... + xk * 2^(bits*k) */
-  mpz_set_ui (x, 0);
-  for (i = 0; i < *nb_curves; i++)
-    {
-      mpz_mul_2exp (x, x, n_bits);
-      mpz_add (x, x, factors[*nb_curves - 1 - i]);
-    }
+  mpz_set_ui (mutable_params->x, 0);
+  /* Equivalent to using mpz_mul_2exp and mpz_add while avoiding O(n*k) limp copies */
+  mpz_realloc2(mutable_params->x, nb_curves * n_bits);
+  for (i = 0; i < nb_curves; i++)
+    for (size_t j = 0; j < n_bits; j++)
+      if (mpz_tstbit (factors[i], j))
+        mpz_setbit(mutable_params->x, j + n_bits * i);
 
   /* was a factor found in stage 1 ? */
   if (youpi != ECM_NO_FACTOR_FOUND)
-      goto end_gpu_ecm_rhotable;
-
-  /* If using 2^k +/-1 modulus and 'nobase2step2' flag is set,
-     set default (-nobase2) modular method and remap P.x, P.y, and P.A */
-  if (modulus->repr == ECM_MOD_BASE2 && nobase2step2)
-    {
-      mpmod_clear (modulus);
-
-      repr = ECM_MOD_NOBASE2;
-      if (mpmod_init (modulus, n, repr) != 0) /* reset modulus for nobase2 */
-        {
-          youpi = ECM_ERROR;
-          goto end_gpu_ecm_rhotable;
-        }
-    }
+      goto end_gpu_ecm_factors;
 
   if (mpz_cmp (B2, B2min) < 0)
-      goto end_gpu_ecm_rhotable;
+      goto end_gpu_ecm_factors;
 
   st2 = cputime ();
-  
+
   P.disc = 0; /* For stage2 this needs to be 0, in order not to use CM stuff */
 
-  for (i = 0; i < *nb_curves; i++)
+  for (i = 0; i < nb_curves; i++)
     {
       /* hack to reduce verbose Step 2 */
-      if (verbose > 0)
-        set_verbose (verbose-1);
+      if (params->verbose > 0)
+        set_verbose (params->verbose-1);
 
-      if (test_verbose (OUTPUT_RESVERBOSE)) 
+      if (test_verbose (OUTPUT_RESVERBOSE))
         outputf (OUTPUT_RESVERBOSE, "x=%Zd\n", factors[i]);
 
-      if (stop_asap != NULL && (*stop_asap) ())
+      if (params->stop_asap != NULL && params->stop_asap())
           goto end_gpu_ecm_rhotable;
-    
+
       mpres_set_z (P.x, factors[i], modulus);
       mpres_set_ui (P.y, 1, modulus);
       A_from_sigma (tmp_A, i+firstsigma_ui, modulus->orig_modulus);
       mpres_set_z (P.A, tmp_A, modulus);
-  
+
       /* compute stage 2 */
       youpi = montgomery_to_weierstrass (factors[i], P.x, P.y, P.A, modulus);
       if (youpi != ECM_NO_FACTOR_FOUND)
@@ -678,16 +456,16 @@ gpu_ecm (mpz_t f, mpz_t x, int param, mpz_t firstsigma, mpz_t n, mpz_t go,
           mpres_get_z (t, P.y, modulus);
           outputf (OUTPUT_RESVERBOSE, ", %Zd)\n", t);
           mpres_get_z (t, P.A, modulus);
-          outputf (OUTPUT_RESVERBOSE, "on curve Y^2 = X^3 + %Zd * X + b\n", 
+          outputf (OUTPUT_RESVERBOSE, "on curve Y^2 = X^3 + %Zd * X + b\n",
                        t);
           mpz_clear (t);
         }
- 
-      youpi = stage2 (factors[i], &P, modulus, dF, k, &root_params, use_ntt, 
-                      TreeFilename, i+1, stop_asap);
-      
+
+      youpi = stage2 (factors[i], &P, modulus, dF, params->k, &root_params,
+                      params->use_ntt, params->TreeFilename, i+1, params->stop_asap);
+
     next_curve:
-      set_verbose (verbose);
+      set_verbose (params->verbose);
 
       if (youpi != ECM_NO_FACTOR_FOUND)
         {
@@ -706,64 +484,63 @@ gpu_ecm (mpz_t f, mpz_t x, int param, mpz_t firstsigma, mpz_t n, mpz_t go,
   youpi = factor_found;
 
   st2 = elltime (st2, cputime ());
-  outputf (OUTPUT_NORMAL, "Computing %u Step 2 on CPU took %ldms\n", 
-                                                              *nb_curves, st2);
-  outputf (OUTPUT_VERBOSE, "Throughput: %.3f Step 2 per second ", 
-                                  1000 * ((double)(*nb_curves))/((double)st2));
-  outputf (OUTPUT_VERBOSE, "(on average %0.2fms per Step 2)\n", 
-                                         ((double) st2)/((double) *nb_curves));
+  outputf (OUTPUT_NORMAL, "Computing %u Step 2 on CPU took %ldms\n",
+                          nb_curves, st2);
+  outputf (OUTPUT_VERBOSE, "Throughput: %.3f Step 2 per second ",
+                           1000.0 * nb_curves /st2);
+  outputf (OUTPUT_VERBOSE, "(on average %0.2fms per Step 2)\n",
+                           ((double) st2)/((double) nb_curves));
   tottime += st2;
 
+end_gpu_ecm_factors:
+
+  reducefactors(factors, array_found, nb_curves);
+
+  /* If f0, ,fk are the factors found (in stage 1 or 2)
+   * f = f0 + f1*n + .. + fk*n^k
+   * The purpose of this construction is to be able to return more than one
+   * factor if needed without breaking the lib interface (as gcd(f,n)=gcd(f0,n).
+   */
+  mpz_set_ui (f, 0);
+  for (i = 0; i < nb_curves; i++)
+  {
+    /* invert order of factors so they are processed in same order found */
+    if (array_found[nb_curves-1-i] != ECM_NO_FACTOR_FOUND)
+      {
+        mpz_mul (f, f, n);
+        mpz_add (f, f, factors[nb_curves-1-i]);
+      }
+  }
+
+  for (i = 0; i < nb_curves; i++)
+      mpz_clear (factors[i]);
+
+  free (array_found);
+  free (factors);
 
 end_gpu_ecm_rhotable:
   if (test_verbose (OUTPUT_VERBOSE))
     {
       if (mpz_cmp_d (B2min, B1) == 0)
         {
-          if (youpi == ECM_NO_FACTOR_FOUND && 
-              (stop_asap == NULL || !(*stop_asap)()))
-              print_exptime (B1, B2, dF, k, root_params.S, 
-                             (long) (tottime / *nb_curves), param);
+          if (youpi == ECM_NO_FACTOR_FOUND &&
+              (params->stop_asap == NULL || !params->stop_asap()))
+              print_exptime (B1, B2, dF, params->k, root_params.S,
+                             (long) (tottime / nb_curves), params->param);
           rhoinit (1, 0); /* Free memory of rhotable */
         }
     }
 
 
-  reducefactors(factors, array_found, *nb_curves);
-
-  /* If f0, ,fk are the factors found (in stage 1 or 2) 
-   * f = f0 + f1*n + .. + fk*n^k
-   * The purpose of this construction is to be able to return more than one
-   * factor if needed without breaking the lib interface (as gcd(f,n)=gcd(f0,n).
-   */
-  mpz_set_ui (f, 0);
-  for (i = 0; i < *nb_curves; i++)
-  {
-    /* invert order of factors so they are processed in same order found */
-    if (array_found[*nb_curves-1-i] != ECM_NO_FACTOR_FOUND)
-      {
-        mpz_mul (f, f, n);
-        mpz_add (f, f, factors[*nb_curves-1-i]);
-      }
-  }
-
 end_gpu_ecm:
   mpz_clear (root_params.i0);
-  mpz_clear (B2);
-  mpz_clear (B2min);
-
-  for (i = 0; i < *nb_curves; i++)
-      mpz_clear (factors[i]);
-
-  free (array_found);
-  free (factors);
-
-end_gpu_ecm2:
-  mpz_clear (tmp_A);
   mpres_clear (P.A, modulus);
   mpres_clear (P.y, modulus);
   mpres_clear (P.x, modulus);
   mpmod_clear (modulus);
+  mpz_clear (tmp_A);
+  mpz_clear (B2);
+  mpz_clear (B2min);
 
   return youpi;
 }
