@@ -278,9 +278,9 @@ extern "C" {
 
 /* getprime.c */
 #define getprime __ECM(getprime)
-double   getprime       ();
+double   getprime       (void);
 #define getprime_clear __ECM(getprime_clear)
-void     getprime_clear ();
+void     getprime_clear (void);
 #define getprime_seek __ECM(getprime_seek)
 void getprime_seek (double);
 
@@ -326,10 +326,11 @@ void duplicate (mpres_t, mpres_t, mpres_t, mpres_t, mpmod_t, mpres_t, mpres_t,
 #define ecm_mul __ECM(ecm_mul)
 void ecm_mul (mpres_t, mpres_t, mpz_t, mpmod_t, mpres_t);
 #define print_B1_B2_poly __ECM(print_B1_B2_poly)
-void print_B1_B2_poly (int, int, double, double, mpz_t, mpz_t, mpz_t, int S,  
-                       mpz_t, int, int, mpz_t, int, unsigned int);
+void print_B1_B2_poly (int, int, double, double, const mpz_t, const mpz_t,
+                       const mpz_t, int S, const mpz_t, int, int, const mpz_t,
+                       int, unsigned int);
 #define set_stage_2_params __ECM(set_stage_2_params)
-int set_stage_2_params (mpz_t, mpz_t, mpz_t, mpz_t, root_params_t *,
+int set_stage_2_params (mpz_t, const mpz_t, mpz_t, const mpz_t, root_params_t *,
                         double, unsigned long *, const int, int, int *,
                         unsigned long *, char *, double, int, mpmod_t);
 #define print_expcurves __ECM(print_expcurves)
@@ -605,7 +606,7 @@ unsigned int F_mul (mpz_t *, mpz_t *, mpz_t *, unsigned int, int,
 unsigned int F_mul_trans (mpz_t *, mpz_t *, mpz_t *, unsigned int,
                           unsigned int, unsigned int, mpz_t *);
 #define F_clear __ECM(F_clear)
-void F_clear ();
+void F_clear (void);
 
 /* rho.c */
 #define rhoinit __ECM(rhoinit)
@@ -678,6 +679,19 @@ unsigned long get_random_ul (void);
 #ifdef HAVE_GWNUM
 int  gw_ecm_stage1 (mpz_t, curve *, mpmod_t, double, double *, mpz_t,
                     double, unsigned long, unsigned long, signed long);
+
+/* We will use gwnum if either
+
+     (1) the input number is of the form k*b^n+c (or is a factor of this number)
+         and either
+         (a) k*b^n+c != 2^n+/-1 and k*b^n+c > 2^gw_knbc_threshold, or
+         (b) k*b^n+c == 2^n+/-1 and n >= gw_spec_div_threshold
+
+  or (2) the input number N is not of the form k*b^n+c, but N > 2^gw_generic_threshold
+*/
+#define GWNUM_KBNC_THRESHOLD 1130 /* bits */
+#define GWNUM_SPEC_DIV_THRESHOLD 1215 /* bits */
+#define GWNUM_GENERIC_THRESHOLD 2100 /* bits */
 #endif
 
 /* batch.c */
