@@ -22,7 +22,6 @@ http://www.gnu.org/licenses/ or write to the Free Software Foundation, Inc.,
 #include <string.h>
 #include <time.h>
 #include "ecm-ecm.h"
-#include "getprime_r.h"
 
 #ifdef HAVE_STRINGS_H
 # include <strings.h> /* for strncasecmp */
@@ -42,13 +41,13 @@ http://www.gnu.org/licenses/ or write to the Free Software Foundation, Inc.,
  *   Unary minus is supported:   -n     -500                     *
  *   Exponentation:              n^m    2^500                    *
  *   Simple factorial:           n!     53!  == 1*2*3*4...*52*53 *
- *   Multi-factorial:            n!m    15!3 == 15.12.9.6.3      *
+ *   Multi-factorial:            n!m    14!3 == 14.11.8.5.2      *
  *   Simple Primorial:           n#     11# == 2*3*5*7*11        *
  *   Reduced Primorial:          n#m    17#5 == 5.7.11.13.17     *
  *                                                               *
  * Supported functions:	(case insensitive)			 *
- *   Phi(n,x)							 *
- *   GCD(m,n)							 *
+ *   Phi(n,x)			Phi(3,5) == 1+x+x^2 = 31	 *
+ *   GCD(m,n)			GCD(120, 28) = 4		 *
  *   U(p,q,n)							 *
  *   primU(p,q,n)						 *
  *   TODO: PhiL(k,n), PhiM(k,n)  				 *
@@ -65,9 +64,9 @@ http://www.gnu.org/licenses/ or write to the Free Software Foundation, Inc.,
 static mpz_t t, mpOne;
 static char *expr_str;
 
-static void eval_power (mpz_t prior_n, mpz_t n,char op);
-static void eval_product (mpz_t prior_n, mpz_t n,char op);
-static void eval_sum (mpz_t prior_n, mpz_t n,char op);
+static void eval_power (mpz_t n, const mpz_t arg, char op);
+static void eval_product (mpz_t n, const mpz_t arg, char op);
+static void eval_sum (mpz_t n, const mpz_t arg, char op);
 static int  eval_Phi (mpz_t *params, mpz_t n);
 static int  eval_PhiL (mpz_t *params, mpz_t n);
 static int  eval_PhiM (mpz_t *params, mpz_t n);
@@ -159,6 +158,9 @@ ChompLine:
 	}
       mpz_clear(t);
     }
+#if defined (DEBUG_EVALUATOR)
+  gmp_fprintf (stderr, "eval(\"%s\") = %Zd\n", expr, n->n);
+#endif
   free(expr);
   return ret;
 }
@@ -219,101 +221,118 @@ JoinLinesLoop:
 	}
       mpz_clear(t);
     }
+#if defined (DEBUG_EVALUATOR)
+  gmp_fprintf (stderr, "eval_str(\"%s\") = %Zd\n", expr, n->n);
+#endif
   free(expr);
   if (EndChar && *EndChar)
     *EndChar = c;
   return ret;
 }
 
-void eval_power (mpz_t prior_n, mpz_t n,char op)
+/**
+ * Evaluate "power" style operations.
+ * Arguments
+ *      n: output also possibly second argument
+ *      arg: first argument to operation
+ *      op: operation to perform, no-op if not a power argument
+ */
+void
+eval_power (mpz_t n, const mpz_t arg, char op)
 {
 #if defined (DEBUG_EVALUATOR)
-  if ('#'==op || '^'==op || '!'==op || '@'==op || '$'==op)
+  if ('^'==op || '!'==op || '@'==op || '#'==op || '$'==op)
     {
-      fprintf (stderr, "eval_power ");
-      mpz_out_str(stderr, 10, prior_n);
-      fprintf (stderr, "%c", op);
-      mpz_out_str(stderr, 10, n);
-      fprintf (stderr, "\n");
+      gmp_fprintf (stderr, "eval power %Zd%c%Zd\n", arg, op, n);
     }
 #endif
 
+  // arg must be positive for everything except power
+  if (mpz_sgn(arg) < 0) {
+    if ('!'==op || '#'==op) {
+      gmp_fprintf (stderr, "\nError - negative argument [%Zd%c]\n", arg, op);
+      exit (EXIT_FAILURE);
+    }
+    if ('@'==op || '$'==op) {
+      gmp_fprintf (stderr, "\nError - negative argument [%Zd%c%Zd]\n", arg, op, n);
+      exit (EXIT_FAILURE);
+    }
+  }
+
+  // second arg must be positive for everything with two args
+  if ('^'==op || '@'==op || '$'==op) {
+    if (mpz_sgn(n) < 0) {
+      gmp_fprintf (stderr, "\nError - negative argument [%Zd%c%Zd]\n", arg, op, n);
+      exit (EXIT_FAILURE);
+    }
+  }
+
   if ('^'==op)
-    mpz_pow_ui(n,prior_n,mpz_get_ui(n));
-  else if ('!'==op)	/* simple factorial  (syntax n!    example: 7! == 1*2*3*4*5*6*7) */
-    mpz_fac_ui(n,mpz_get_ui(n));
-  else if ('@'==op)	/* Multi factorial   (syntax n!prior_n
-                           Example: 15!3 == 15*12*9*6*3
-                           Note: 15!3 is substituted into 15@3 by the parser */
+    mpz_pow_ui(n, arg, mpz_get_ui(n));
+  else if ('!'==op)	/* simple factorial  (syntax arg!    example: 7! == 1*2*3*4*5*6*7) */
+    mpz_fac_ui(n, mpz_get_ui(arg));
+  else if ('@'==op)	/* Multi factorial   (syntax arg!n
+                           Example: 14!3 == 14*11*8*5*2
+                           Note: 14!3 is substituted into 14@3 by the parser */
     {
-      long nCur;
-      unsigned long nDecr;
-      nCur = mpz_get_si(prior_n);
-      nDecr = mpz_get_ui(n);
-      mpz_set_ui(n,1);
-      while (nCur > 1)
-	{
-	  /* This could be done much more efficiently (bunching mults using smaller "built-ins"), but I am not going to bother for now */
-	  mpz_mul_ui(n,n,nCur);
-	  nCur -= nDecr;
-	}
-    }
-  else if ('#'==op)  /* simple primorial (syntax  n#   example:  11# == 2*3*5*7*11 */
-    {
-      unsigned long nMax;
-      unsigned long p;
-      prime_info_t prime_info;
-
-      prime_info_init (prime_info);
-      nMax = mpz_get_ui (n);
+      /* gmp_printf ("Multi factorial  %Zd!%Zd\n", arg, n); */
+      ASSERT_ALWAYS (mpz_cmp(arg, n) >= 0); // arg >= n
+      mpz_t p, dec;
+      mpz_init_set (p, arg);
+      mpz_init_set (dec, n);
       mpz_set_ui (n, 1);
-      for (p = 2; p <= nMax; p = getprime_mt (prime_info))
-	/* This could be done much more efficiently (bunching mults using smaller "built-ins"), but I am not going to bother for now */
-	mpz_mul_ui (n, n, p);
-      prime_info_clear (prime_info); /* free the prime table */
-    }
-  else if ('$'==op)  /* reduced primorial (syntax  n#prior_n   example:  13#5 == (5*7*11*13) */
-    {
-      unsigned long p;
-      unsigned long nMax;
-      unsigned long nStart;
-      prime_info_t prime_info;
-
-      prime_info_init (prime_info);
-      nMax = mpz_get_ui (prior_n);
-      nStart = mpz_get_ui (n);
-      mpz_set_ui (n, 1);
-      /*printf ("Reduced-primorial  %ld#%ld\n", nMax, nStart);*/
-      for (p = 2; p <= nMax; p = getprime_mt (prime_info))
+      while (mpz_cmp_ui(p, 1) > 0)
 	{
-	  if (p >= nStart)
-	    /* This could be done much more efficiently (bunching mults using smaller "built-ins"), but I am not going to bother for now */
-	    mpz_mul_ui (n, n, p);
+	  /* Could use factor-tree, not worth the extra code. */
+	  mpz_mul (n, n, p);
+	  mpz_sub (p, p, dec);
 	}
-      prime_info_clear (prime_info); /* free the prime table */
+      mpz_clear(p);
+      mpz_clear(dec);
+    }
+  else if ('#'==op)  /* simple primorial (syntax  arg#   example:  11# == 2*3*5*7*11 */
+    {
+      mpz_primorial_ui(n, mpz_get_ui(arg));
+    }
+  else if ('$'==op)  /* reduced primorial (syntax  arg#n   example:  13#5 == (5*7*11*13) */
+    {
+      /* gmp_printf ("Reduced-primorial  %Zd#%Zd\n", arg, n); */
+      ASSERT_ALWAYS (mpz_cmp(arg, n) >= 0); // arg >= n
+      mpz_t p;
+      mpz_init_set(p, n);
+      mpz_set_ui (n, 1);
+      mpz_sub_ui(p, p, 1);
+      mpz_nextprime(p, p);
+      for (; mpz_cmp(p, arg) <= 0; mpz_nextprime(p, p))
+	  /* Could use factor-tree, not worth the extra code. */
+	  mpz_mul (n, n, p);
+      mpz_clear(p);
     }
 }
 
+/**
+ * Evaluate "product" style operations.
+ * Arguments
+ *      n: output also possibly second argument
+ *      arg: first argument to operation
+ *      op: operation to perform, no-op if not a power argument
+ */
 void
-eval_product (mpz_t prior_n, mpz_t n, char op)
+eval_product (mpz_t n, const mpz_t arg, char op)
 {
 #if defined (DEBUG_EVALUATOR)
   if ('*'==op || '.'==op || '/'==op || '%'==op)
     {
-      fprintf (stderr, "eval_product ");
-      mpz_out_str(stderr, 10, prior_n);
-      fprintf (stderr, "%c", op);
-      mpz_out_str(stderr, 10, n);
-      fprintf (stderr, "\n");
+      gmp_fprintf (stderr, "eval_product %Zd%c%Zd\n", arg, op, n);
     }
 #endif
   if ('*' == op || '.' == op)
-    mpz_mul (n, prior_n, n);
+    mpz_mul (n, arg, n);
   else if ('/' == op)
     {
       mpz_t r;
       mpz_init (r);
-      mpz_tdiv_qr (n, r, prior_n, n);
+      mpz_tdiv_qr (n, r, arg, n);
       if (mpz_cmp_ui (r, 0) != 0)
         {
           fprintf (stderr, "Parsing Error: inexact division\n");
@@ -322,26 +341,30 @@ eval_product (mpz_t prior_n, mpz_t n, char op)
       mpz_clear (r);
     }
   else if ('%' == op)
-    mpz_tdiv_r (n, prior_n, n);
+    mpz_tdiv_r (n, arg, n);
 }
 
-void eval_sum (mpz_t prior_n, mpz_t n,char op)
+/**
+ * Evaluate "sum" style operations.
+ * Arguments
+ *      n: output also possibly second argument
+ *      arg: first argument to operation
+ *      op: operation to perform, no-op if not a power argument
+ */
+void
+eval_sum (mpz_t n, const mpz_t arg, char op)
 {
 #if defined (DEBUG_EVALUATOR)
   if ('+'==op || '-'==op)
     {
-      fprintf (stderr, "eval_sum ");
-      mpz_out_str(stderr, 10, prior_n);
-      fprintf (stderr, "%c", op);
-      mpz_out_str(stderr, 10, n);
-      fprintf (stderr, "\n");
+      gmp_fprintf (stderr, "eval_sum %Zd%c%Zd\n", arg, op, n);
     }
 #endif
 
   if ('+' == op)
-    mpz_add(n,prior_n,n);
+    mpz_add(n, arg, n);
   else if ('-' == op)
-    mpz_sub(n,prior_n,n);
+    mpz_sub(n, arg, n);
 }
 
 int eval_Phi (mpz_t* params, mpz_t n)
@@ -352,7 +375,6 @@ int eval_Phi (mpz_t* params, mpz_t n)
   unsigned long B;
   unsigned long p;
   mpz_t D, T, org_n;
-  prime_info_t prime_info;
   
   /* deal with trivial cases first */
   if (mpz_cmp_ui (params[0], 0) == 0)
@@ -409,8 +431,7 @@ int eval_Phi (mpz_t* params, mpz_t n)
   B = mpz_get_ui (params[0]);
 
   /* Obtain the factors of B */
-  prime_info_init (prime_info);
-  for (p = 2; p <= B; p = getprime_mt (prime_info))
+  for (p = 2; p <= B; p += 1 + (p>2))
     {
       if (B % p == 0)
 	{
@@ -420,14 +441,13 @@ int eval_Phi (mpz_t* params, mpz_t n)
 	  do { B /= p; } while (B % p == 0);
         }
      }
-  prime_info_clear (prime_info); /* free the prime tables */
   B = mpz_get_si (params[0]);
 
   mpz_init_set (org_n, n);
   mpz_set_ui (n, 1);
   mpz_init_set_ui (D, 1);
   mpz_init (T);
-	      
+
   for(dw=0;(dw<(1U<<dwFactors)); dw++)
     {
       /* for all Mobius terms */
@@ -435,7 +455,7 @@ int eval_Phi (mpz_t* params, mpz_t n)
       int iMobius=0;
       unsigned dwIndex=0;
       unsigned dwMask=1;
-		  
+
       while(dwIndex < dwFactors)
 	{
 	  if(dw&dwMask)
@@ -448,34 +468,18 @@ int eval_Phi (mpz_t* params, mpz_t n)
 	  dwMask<<=1;
 	  ++dwIndex;
 	}
-      /*
-      fprintf (stderr, "Taking ");
-      mpz_out_str(stderr, 10, org_n);
-      fprintf (stderr, "^%d-1\n", iPower);
-      */
+      // gmp_fprintf (stderr, "Taking %Zd^%d-1\n", org_n, iPower);
       mpz_pow_ui(T, org_n, iPower);
       mpz_sub_ui(T, T, 1);
     
       if(iMobius&1)
       {
-	/*
-	fprintf (stderr, "Muling D=D*T  ");
-	mpz_out_str(stderr, 10, D);
-	fprintf (stderr, "*");
-	mpz_out_str(stderr, 10, T);
-	fprintf (stderr, "\n");
-	*/
+	// gmp_fprintf (stderr, "Muling D=D*T  %Zd*%Zd\n", D, T);
 	mpz_mul(D, D, T);
       }
       else
       {
-	/*
-	fprintf (stderr, "Muling n=n*T  ");
-	mpz_out_str(stderr, 10, n);
-	fprintf (stderr, "*");
-	mpz_out_str(stderr, 10, T);
-	fprintf (stderr, "\n");
-	*/
+	// gmp_fprintf (stderr, "Muling n=n*T  %Zd*%Zd\n", n, T);
 	mpz_mul(n, n, T); 
       }
   }
@@ -579,7 +583,7 @@ int aurif (mpz_t output, mpz_t n, mpz_t base, int sign) // Evaluate Aurifeullian
     }
   mpz_set(output, C);
   (sign>0 ? mpz_addmul : mpz_submul)(output, D, l);
-//  gmp_fprintf(stderr, "Calculated base=%Zd, exp=%Zd, C=%Zd, D=%Zd, output=%Zd\n",base,n,C,D,output);
+  // gmp_fprintf (stderr, "Calculated base=%Zd, exp=%Zd, C=%Zd, D=%Zd, output=%Zd\n",base,n,C,D,output);
   mpz_clears(orig_base,C,D,l,m,NULL);
   return 1;
 }
@@ -683,7 +687,7 @@ int eval_U (mpz_t *params, mpz_t n)
 	    mpz_addmul (U1,U0,params[1]); // U0 is 2k-1, U1 is 2k
 	    mpz_divexact (U1,U1,params[0]);
          }
-	 /* gmp_printf("%d %Zd %Zd\n",k,U0,U1); */
+	 /* gmp_fprintf (stderr, "%d %Zd %Zd\n",k,U0,U1); */
     }
   mpz_set(n, U1);
 
@@ -774,9 +778,7 @@ int eval_primU (mpz_t* params, mpz_t n)
 	  dwMask<<=1;
 	  ++dwIndex;
 	}
- /*     
-      gmp_fprintf (stderr, "Taking U(%Zd,%Zd,%d)\n", P,Q,iPower);
- */     
+      // gmp_fprintf (stderr, "Taking U(%Zd,%Zd,%d)\n", P,Q,iPower);
       mpz_set_ui(T,iPower);
       if(eval_U(params, T)==0)
       {
@@ -785,24 +787,12 @@ int eval_primU (mpz_t* params, mpz_t n)
     
       if(iMobius&1)
       {
-/*	
-	fprintf (stderr, "Muling D=D*T  ");
-	mpz_out_str(stderr, 10, D);
-	fprintf (stderr, "*");
-	mpz_out_str(stderr, 10, T);
-	fprintf (stderr, "\n");
-*/	
+	// gmp_fprintf (stderr, "Muling D=D*T  %Zd*%Zd\n", D, T);
 	mpz_mul(D, D, T);
       }
       else
       {
-/*	
-	fprintf (stderr, "Muling n=n*T  ");
-	mpz_out_str(stderr, 10, n);
-	fprintf (stderr, "*");
-	mpz_out_str(stderr, 10, T);
-	fprintf (stderr, "\n");
-*/	
+	// gmp_fprintf (stderr, "Muling n=n*T  %Zd*%Zd\n", n, T);
 	mpz_mul(n, n, T); 
       }
   }
@@ -939,13 +929,14 @@ MONADIC_SUFFIX_LOOP:
 	    
       if (0==op || ')'==op || ']'==op || '}'==op || (','==op&&bInFuncParams))
 	{
-	  eval_power (n_stack[2],n,op_stack[2]);
-	  eval_product (n_stack[1],n,op_stack[1]);
-	  eval_sum (n_stack[0],n,op_stack[0]);
+	  eval_power (n, n_stack[2], op_stack[2]);
+	  eval_product (n, n_stack[1], op_stack[1]);
+	  eval_sum (n, n_stack[0], op_stack[0]);
 	  mpz_set(t, n);
 	  mpz_clear(n);
 	  for (i=0;i<5;i++)
             {
+              op_stack[i]=0;
               mpz_clear(n_stack[i]);
               mpz_clear(param_stack[i]);
             }
@@ -956,7 +947,7 @@ MONADIC_SUFFIX_LOOP:
 	{
 	  if ('^' == op)
 	    {
-	      eval_power (n_stack[2],n,op_stack[2]);
+	      eval_power (n, n_stack[2], op_stack[2]);
 	      mpz_set(n_stack[2], n);
 	      op_stack[2]='^';
 	    }
@@ -969,7 +960,7 @@ MONADIC_SUFFIX_LOOP:
 		  op_stack[2]='!';
 		  goto MONADIC_SUFFIX_LOOP;
 		}
-	      eval_power (n_stack[2],n,op_stack[2]);
+	      eval_power (n, n_stack[2], op_stack[2]);
 	      mpz_set(n_stack[2], n);
 	      op_stack[2]='@';
 	    }
@@ -982,7 +973,7 @@ MONADIC_SUFFIX_LOOP:
 		  op_stack[2]='#';
 		  goto MONADIC_SUFFIX_LOOP;
 		}
-	      eval_power (n_stack[2],n,op_stack[2]);
+	      eval_power (n, n_stack[2], op_stack[2]);
 	      mpz_set(n_stack[2], n);
 	      op_stack[2]='$';
 	    }
@@ -990,9 +981,9 @@ MONADIC_SUFFIX_LOOP:
 	    {
 	      if ('.'==op || '*'==op || '/'==op || '%'==op)
 		{
-		  eval_power (n_stack[2],n,op_stack[2]);
+		  eval_power (n, n_stack[2], op_stack[2]);
 		  op_stack[2]=0;
-		  eval_product (n_stack[1],n,op_stack[1]);
+		  eval_product (n, n_stack[1], op_stack[1]);
 		  mpz_set(n_stack[1], n);
 		  op_stack[1]=op;
 		}
@@ -1000,11 +991,11 @@ MONADIC_SUFFIX_LOOP:
 		{
 		  if ('+'==op || '-'==op)
 		    {
-		      eval_power (n_stack[2],n,op_stack[2]);
+		      eval_power (n, n_stack[2], op_stack[2]);
 		      op_stack[2]=0;
-		      eval_product (n_stack[1],n,op_stack[1]);
+		      eval_product (n, n_stack[1], op_stack[1]);
 		      op_stack[1]=0;
-		      eval_sum (n_stack[0],n,op_stack[0]);
+		      eval_sum (n, n_stack[0], op_stack[0]);
 		      mpz_set(n_stack[0], n);
 		      op_stack[0]=op;
 		    }
