@@ -79,7 +79,7 @@ signal_handler (int sig)
 }
 
 int
-stop_asap_test ()
+stop_asap_test (void)
 {
   return exit_asap_value;
 }
@@ -127,7 +127,6 @@ usage (void)
     printf ("  -maxmem n    use at most n MB of memory in stage 2\n");
     printf ("  -stage1time n add n seconds to ECM stage 1 time (for expected time est.)\n");
 
-    printf ("  -I f         increment B1 by f*sqrt(B1) on each run\n");
     printf ("  -inp file    Use file as input (instead of redirecting stdin)\n");
     printf ("  -one         Stop processing a candidate if a factor is found (looping mode)\n");
     printf ("  -go val      Preload with group order val, which can be a simple expression,\n");
@@ -137,21 +136,22 @@ usage (void)
     printf ("  -bsaves file With -param 1-3, save stage 1 exponent in file.\n");
     printf ("  -bloads file With -param 1-3, load stage 1 exponent from file.\n");
 #ifdef WITH_GPU
-    printf ("  -gpu         Use GPU-ECM for stage 1.\n");
-#if HAVE_CGBN_H
-    printf ("  -cgbn        Use CGBN for GPU-ECM stage 1 computation.\n");
-#endif /* HAVE_CGBN_H */
+    printf ("  -gpu         Use CGBN for computations stage 1.\n");
     printf ("  -gpudevice n Use device n to execute GPU code (by default, "
                                                           "CUDA chooses)\n");
     printf ("  -gpucurves n Compute on n curves in parallel on the GPU (by "
                                                   "default, CUDA chooses)\n");
 #endif /* WITH_GPU */
+#ifdef HAVE_GWNUM
+    printf ("  -force-gwnum       Use gwnum routines for ECM stage 1 (if possible).\n");
+    printf ("  -force-no-gwnum    Do not use gwnum routines for ECM stage 1.\n");
+#endif /* HAVE_GWNUM */
     printf ("  -h, --help   Prints this help and exit.\n");
 }
 
 /* Print parameters that were used to build GMP-ECM */
 static void
-print_config ()
+print_config (void)
 {
   printf ("Compilation options:\n");
 #ifdef __MPIR_VERSION
@@ -352,7 +352,7 @@ int
 main (int argc, char *argv[])
 {
   char **argv0 = argv;
-  mpz_t x, y, sigma, A, f, orig_x0, orig_y0, B2, B2min, startingB2min, tmp_n;
+  mpz_t x, y, sigma, A, f, orig_x0, orig_y0, B2, B2min, startingB2min, orig_n;
   mpcandi_t n;
   mpgocandi_t go;
   mpq_t rat_x0, rat_y0, rat_A;
@@ -388,7 +388,7 @@ main (int argc, char *argv[])
         store the last number processed and the factors found for this it */
   int resume_wasPrp = 0; /* 1 if resume_lastN/resume_lastfac is a PRP */
   int primetest = 0, saveappend = 0;
-  double autoincrementB1 = 0.0, startingB1;
+  double startingB1;
   unsigned int count = 1; /* number of curves for each number */
   unsigned int cnt = 0;   /* number of remaining curves for current number */
   int deep=1;
@@ -403,9 +403,9 @@ main (int argc, char *argv[])
   unsigned long gw_b = 0;  /* set default values for gwnum poly k*b^n+c */
   unsigned long gw_n = 0;  /* set default values for gwnum poly k*b^n+c */
   signed long gw_c = 0;    /* set default values for gwnum poly k*b^n+c */
+  signed long gw_cl_flag = 0;
 #endif
   int use_gpu = 0; /* Do we use the GPU for stage 1 (by default no) */
-  int gpucgbn = 0; /* Do we use CGBN for stage 1 GPU computation (by default no) */
   int gpudevice = -1; /* Which device do we use for GPU code (by default CUDA */
                       /* chooses)                                             */
   unsigned int gpucurves = 0; /* How many curves do we want for GPU code */ 
@@ -509,6 +509,23 @@ main (int argc, char *argv[])
 	  argv++;
 	  argc--;
 	}
+#ifdef HAVE_GWNUM
+      /* switches to facilitate finding crossovers or force faster method  */
+      else if (strcmp (argv[1], "-force-gwnum") == 0)
+	{
+	  gw_cl_flag = 1.0; /* always use gwnum if input number > 2^350 */
+	  argv++;
+	  argc--;
+	}
+      else if (strcmp (argv[1], "-force-no-gwnum") == 0)
+	{
+	  gw_cl_flag = -1.0; /* Do not use gwnum */
+	  argv++;
+	  argc--;
+	}
+      /* note that gw_cl_flag == 0.0 will indicate that neither choice above
+         was specified and the estimated crossovers in ecm-impl.h will be used */
+#endif /* HAVE_GWNUM */
       else if (strcmp (argv[1], "-primetest") == 0)
         {
           primetest = 1;
@@ -720,17 +737,6 @@ main (int argc, char *argv[])
 	  argv += 2;
 	  argc -= 2;
 	}
-      else if ((argc > 2) && (strcmp (argv[1], "-I") == 0))
-	{
-	  autoincrementB1 = strtod (argv[2], NULL);
-	  if (autoincrementB1 <= 0.0)
-	    {
-	      fprintf (stderr, "Error, the -I f option requires f > 0\n");
-	      exit (EXIT_FAILURE);
-  	    }
-	  argv += 2;
-	  argc -= 2;
-	}
       else if ((argc > 2) && (strcmp (argv[1], "-inp") == 0))
 	{
 	  infilename = argv[2];
@@ -788,23 +794,12 @@ main (int argc, char *argv[])
 	  argc -= 2;
 	}
 #ifdef WITH_GPU
-      else if (strcmp (argv[1], "-gpu") == 0)
+      else if (strcmp (argv[1], "-gpu") == 0 || strcmp (argv[1], "-cgbn") == 0)
         {
           use_gpu = 1;
           argv++;
           argc--;
         }
-      else if (strcmp (argv[1], "-cgbn") == 0)
-        {
-          use_gpu = 1;
-          gpucgbn = 1;
-#ifndef HAVE_CGBN_H
-          fprintf (stderr, "CGBN not present; configure with --with-cgbn-include\n");
-          exit (EXIT_FAILURE);
-#endif /* !HAVE_CGBN_H */
-          argv++;
-          argc--;
-        }
       else if ((argc > 2) && (strcmp (argv[1], "-gpudevice") == 0))
         {
           gpudevice = atoi (argv[2]);
@@ -1046,7 +1041,6 @@ main (int argc, char *argv[])
   params->maxmem = maxmem;
   params->stage1time = stage1time;
   params->gpu = use_gpu;   /* If WITH_GPU is not defined it will always be 0 */
-  params->gpu_cgbn = gpucgbn; /* If !HAVE_CGBN_H will always be 0 */
   params->gpu_device = gpudevice; /* If WITH_GPU is not defined or    */
                                   /* use_gpu = 0, it has no meaning   */
   params->gpu_number_of_curves = gpucurves; /* If WITH_GPU is not defined or */
@@ -1372,7 +1366,9 @@ main (int argc, char *argv[])
         {
           if (!IS_BATCH_MODE(param))
             {
-              fprintf (stderr, "Error, -bsaves makes sense in batch mode only\n");
+              fprintf (stderr, "Error, -bsaves makes sense in batch mode only "
+                       "(-param in {%d,%d,%d})\n", ECM_PARAM_BATCH_SQUARE,
+                       ECM_PARAM_BATCH_2, ECM_PARAM_BATCH_32BITS_D);
               exit (EXIT_FAILURE);
             }
           /* if batch_s <> 1, it means it was already initialized,
@@ -1387,7 +1383,9 @@ main (int argc, char *argv[])
           int st = cputime ();
           if (!IS_BATCH_MODE(param))
             {
-              fprintf (stderr, "Error, -bloads makes sense in batch mode only\n");
+              fprintf (stderr, "Error, -bloads makes sense in batch mode only "
+                       "(-param in {%d,%d,%d})\n", ECM_PARAM_BATCH_SQUARE,
+                       ECM_PARAM_BATCH_2, ECM_PARAM_BATCH_32BITS_D);
               exit (EXIT_FAILURE);
             }
           params->batch_last_B1_used = B1;
@@ -1470,8 +1468,20 @@ main (int argc, char *argv[])
         params->use_ntt = use_ntt;
 
 #ifdef HAVE_GWNUM
+      params->gw_cl_flag = gw_cl_flag; /* save */
+
+      /* check if the input number is too small (2^350) for gwnum */
+      if (mpz_sizeinbase(n.n, 2) < 350)
+      {
+        params->gw_b = 0;
+        params->gw_k = 0.0; /* indicate that input is too small for gwnum */
+        if (verbose > OUTPUT_NORMAL)
+          printf ("Input number is too small for gwnum\n");
+      }
+      else
+      {
       /* check if the input number can be represented as k*b^n+c */
-      if (kbnc_z (&gw_k, &gw_b, &gw_n, &gw_c, n.n))
+      if (kbnc_str (&gw_k, &gw_b, &gw_n, &gw_c, n.cpExpr, n.n))
         {
           params->gw_k = gw_k;
           params->gw_b = gw_b;
@@ -1481,7 +1491,7 @@ main (int argc, char *argv[])
             printf ("Found number: %.0f*%lu^%lu + %ld\n",
                     gw_k, gw_b, gw_n, gw_c);
         }
-      else if (kbnc_str (&gw_k, &gw_b, &gw_n, &gw_c, n.cpExpr, n.n))
+      else if (kbnc_z (&gw_k, &gw_b, &gw_n, &gw_c, n.n))
         {
           params->gw_k = gw_k;
           params->gw_b = gw_b;
@@ -1493,9 +1503,9 @@ main (int argc, char *argv[])
         }
       else
         {
-          if (verbose > OUTPUT_NORMAL)
-            printf ("Did not find a gwnum poly for the input number.\n");
+          params->gw_b = 0;
         }
+      }
 #endif
 
       if (timestamp)
@@ -1517,10 +1527,11 @@ main (int argc, char *argv[])
         /* if torsion was used, some factor may have been found... */
         result = ecm_factor (f, n.n, B1, params);
 
-      if (result == ECM_ERROR)
+      if (result == ECM_ERROR || result == ECM_USER_ERROR)
         {
-          fprintf (stderr, "Please report internal errors at <%s>.\n",
-                   PACKAGE_BUGREPORT);
+          if (result == ECM_ERROR) // internal error
+            fprintf (stderr, "Please report internal errors at <%s>.\n",
+                     PACKAGE_BUGREPORT);
           exit (EXIT_FAILURE);
         }
       
@@ -1534,9 +1545,9 @@ main (int argc, char *argv[])
               cnt -= params->gpu_number_of_curves; 
         }
 
-      /* When GPU is used we need to have the value of N before it is 
+      /* When GPU is used we need to have the value of N before it is
         divided by potential factor in f */
-      mpz_init_set (tmp_n, n.n);
+      mpz_init_set (orig_n, n.n);
 
       if (result != ECM_NO_FACTOR_FOUND)
         {
@@ -1547,7 +1558,7 @@ main (int argc, char *argv[])
             {
               if (params->gpu)
                   /* gpu returns multiple factors as f = f0 + f1*n + ... + fk*n^k */
-                  mpz_fdiv_qr (f, tmp_factor, f, tmp_n);
+                  mpz_fdiv_qr (f, tmp_factor, f, orig_n);
               else
                   mpz_set (tmp_factor, f);
 
@@ -1575,11 +1586,11 @@ main (int argc, char *argv[])
       if (savefilename != NULL && !n.isPrp)
         {
         /* TODO Deal with return code */
-	    write_resumefile (savefilename, method, tmp_n, params, &n, 
-			      orig_x0, orig_y0, comment);
+	    write_resumefile (savefilename, method, params, &n,
+			      orig_n, orig_x0, orig_y0, comment);
         }
 
-      mpz_clear (tmp_n);
+      mpz_clear (orig_n);
 
       /* Save the batch exponent s if requested */
       if (savefile_s != NULL)
@@ -1590,16 +1601,6 @@ main (int argc, char *argv[])
                       savefile_s);
         }
 
-      /* advance B1, if autoincrement value had been set during command line 
-         parsing */
-      if (autoincrementB1 > 0.0)
-        {
-          double NewB1;
-          NewB1 = calc_B1_AutoIncrement (B1, autoincrementB1);
-          if (mpz_cmp_d (B2min, B1) <= 0) /* <= might be better than == */
-              mpz_set_d (B2min, NewB1);
-          B1 = NewB1;
-        }
     } /* end of main loop */
 
   free_all1:
